import '../fenwick_tree.dart';

/// Use this when you have a group.
class GroupalFenwickTreeImpl<T> with MonoidalFenwickTreeMixin<T>, GroupalFenwickTreeMixin<T> {
  @override
  final FenwickTreeGroup<T> algebra;
  @override
  final List<T> array;

  GroupalFenwickTreeImpl({
    required final this.algebra,
    required final int size,
  }) : array = List.filled(
          size + 1,
          algebra.identity(),
        );
}

/// Use this when you only have a monoid.
class MonoidalFenwickTreeImpl<T> with MonoidalFenwickTreeMixin<T> {
  @override
  final FenwickTreeMonoid<T> algebra;
  @override
  final List<T> array;

  MonoidalFenwickTreeImpl({
    required final this.algebra,
    required final int size,
  }) : array = List.filled(
          size + 1,
          algebra.identity(),
        );
}

mixin GroupalFenwickTreeMixin<T> implements GroupalFenwickTree<T>, FenwickTreeMemory<T> {
  FenwickTreeGroup<T> get algebra;

  @override
  T range_sum({
    required final int index_from,
    required final int index_to,
  }) {
    final a = index_from;
    final b = index_to;
    assert(b > a, "");
    assert(a > -1, "");
    assert(b > -1, "");
    assert(array.length > a + 1, "");
    assert(array.length > b, "");
    if (a == 0) {
      return sum(
        index_to: b - 1,
      );
    } else {
      return algebra.inverse(
        sum(index_to: b - 1),
        sum(index_to: a - 1),
      );
    }
  }

  @override
  void update({
    required final int index,
    required final T value,
  }) {
    final current_sum = range_sum(
      index_from: index,
      index_to: index + 1,
    );
    final updated_value = algebra.inverse(
      value,
      current_sum,
    );
    increase(
      index: index,
      value: updated_value,
    );
  }
}

mixin MonoidalFenwickTreeMixin<T> implements MonoidalFenwickTree<T>, FenwickTreeMemory<T> {
  FenwickTreeMonoid<T> get algebra;

  @override
  T sum({
    required final int index_to,
  }) {
    int i = index_to;
    assert(i > -1 && array.length > i + 1, "");
    T sum = algebra.identity();
    i++;
    while (i > 0) {
      sum = algebra.operate(sum, array[i]);
      i -= i & -i;
    }
    if (i == array.length - 1) {
      return algebra.operate(sum, array[i]);
    } else {
      return sum;
    }
  }

  @override
  void increase({
    required final int index,
    required final T value,
  }) {
    int i = index;
    assert(i > -1, "");
    i++;
    while (i < array.length) {
      array[i] = algebra.operate(
        array[i],
        value,
      );
      i += i & -i;
    }
  }

  @override
  int length() => array.length - 1;
}

abstract class FenwickTreeMemory<T> {
  List<T> get array;
}
