import 'package:fenwick_tree/fenwick_tree.dart';
import 'package:test/test.dart';

void main() {
  group("$GroupalFenwickTree", () {
    test("smoke test", () {
      final tree = GroupalFenwickTreeImpl<double>(
        algebra: const FenwickTreeMonoidDoublesumImpl(),
        size: 3,
      );
      tree.increase(index: 0, value: 1);
      tree.increase(index: 1, value: 2);
      tree.increase(index: 2, value: 3);
      expect(tree.sum(index_to: 2), 6);
    });
    test("length", () {
      final sut = GroupalFenwickTreeImpl(
        algebra: const FenwickTreeMonoidIntsumImpl(),
        size: 100,
      );
      List.generate(100, (final a) => a).forEach(
        (final a) => sut.increase(
          index: a,
          value: 1,
        ),
      );
      expect(sut.length(), 100);
    });
    // https://www.geeksforgeeks.org/binary-indexed-tree-or-fenwick-tree-2/
    test("increase", () {
      final sut = GroupalFenwickTreeImpl(
        algebra: const FenwickTreeMonoidIntsumImpl(),
        size: 12,
      );
      [2, 1, 1, 3, 2, 3, 4, 5, 6, 7, 8, 9].asMap().entries.forEach(
            (final entry) => sut.increase(
              index: entry.key,
              value: entry.value,
            ),
          );
      expect(
        sut.array,
        [0, 2, 3, 1, 7, 2, 5, 4, 21, 6, 13, 8, 30],
      );
    });
    test("update", () {
      const first = 1.0;
      const second = 10.0;
      const third = 100.0;
      const new_value = 1000.0;
      GroupalFenwickTree<double> make() {
        final tree = GroupalFenwickTreeImpl<double>(
          algebra: const FenwickTreeMonoidDoublesumImpl(),
          size: 3,
        );
        tree.increase(index: 0, value: first);
        tree.increase(index: 1, value: second);
        tree.increase(index: 2, value: third);
        return tree;
      }
      {
        final tree = make();
        expect(tree.sum(index_to: 0), first);
        tree.update(index: 0, value: new_value);
        expect(tree.sum(index_to: 2), new_value + second + third);
      }
      {
        final tree = make();
        expect(tree.sum(index_to: 1), first + second);
        tree.update(index: 1, value: new_value);
        expect(tree.sum(index_to: 2), first + new_value + third);
      }
      {
        final tree = make();
        expect(tree.sum(index_to: 2), first + second + third);
        tree.update(index: 2, value: new_value);
        expect(tree.sum(index_to: 2), first + second + new_value);
      }
    });
    group("sum", () {
      final sut = GroupalFenwickTreeImpl(
        algebra: const FenwickTreeMonoidIntsumImpl(),
        size: 5,
      );
      [1, 2, 3, 4, 5].asMap().entries.forEach(
            (final entry) => sut.increase(
              index: entry.key,
              value: entry.value,
            ),
          );
      test("0", () {
        expect(sut.sum(index_to: 0), 1);
      });
      test("1", () {
        expect(sut.sum(index_to: 1), 1 + 2);
      });
      test("4", () {
        expect(sut.sum(index_to: 4), 1 + 2 + 3 + 4 + 5);
      });
      test("5", () {
        expect(() => sut.sum(index_to: 5), throwsA(const TypeMatcher<AssertionError>()));
      });
    });
    group("rangeSum", () {
      final sut = GroupalFenwickTreeImpl(
        algebra: const FenwickTreeMonoidIntsumImpl(),
        size: 5,
      );
      [1, 2, 3, 4, 5].asMap().entries.forEach(
            (final entry) => sut.increase(
              index: entry.key,
              value: entry.value,
            ),
          );
      test("[0, 5]", () => expect(sut.range_sum(index_from: 0, index_to: 5), 15));
      test("[0, 2]", () => expect(sut.range_sum(index_from: 0, index_to: 2), 3));
      test("[1, 4]", () => expect(sut.range_sum(index_from: 1, index_to: 4), 9));
      test("[1, 5]", () => expect(sut.range_sum(index_from: 1, index_to: 5), 14));
      test(
        "[0, 0]",
        () => expect(
          () => sut.range_sum(index_from: 0, index_to: 0),
          throwsA(const TypeMatcher<AssertionError>()),
        ),
      );
      test(
        "[5, 5]",
        () => expect(
          () => sut.range_sum(index_from: 5, index_to: 5),
          throwsA(const TypeMatcher<AssertionError>()),
        ),
      );
      test(
        "[3, 2]",
        () => expect(
          () => sut.range_sum(index_from: 3, index_to: 2),
          throwsA(const TypeMatcher<AssertionError>()),
        ),
      );
      test(
        "[-1, 2]",
        () => expect(
          () => sut.range_sum(index_from: -1, index_to: 2),
          throwsA(const TypeMatcher<AssertionError>()),
        ),
      );
      test(
        "[0, 6]",
        () => expect(
          () => sut.range_sum(index_from: 0, index_to: 6),
          throwsA(const TypeMatcher<AssertionError>()),
        ),
      );
    });
  });
}

class FenwickTreeMonoidDoublesumImpl implements FenwickTreeGroup<double> {
  const FenwickTreeMonoidDoublesumImpl();

  @override
  double identity() => 0.0;

  @override
  double inverse(
    final double a,
    final double b,
  ) =>
      a - b;

  @override
  double operate(
    final double a,
    final double b,
  ) =>
      a + b;
}

class FenwickTreeMonoidIntsumImpl implements FenwickTreeGroup<int> {
  const FenwickTreeMonoidIntsumImpl();

  @override
  int identity() => 0;

  @override
  int inverse(
    final int a,
    final int b,
  ) =>
      a - b;

  @override
  int operate(
    final int a,
    final int b,
  ) =>
      a + b;
}
